------------------------------------------------------------------------------------------------------------------
--Imports
----------------------------------------------------------------------------------------------------

local errorHandler = require("scenario/intersection/handlers/errorHandler")
local classHelper  = require("scenario/intersection/classHelper")

-------------------------------------------------------------------------------------------------------

local ClassTimer = classHelper.inheritsFrom(nil)

----------------------------------------------------------------------------------------------------
--Instance Methods
----------------------------------------------------------------------------------------------------

function ClassTimer:setActive(active)
    self.active = active
end

function ClassTimer:hasFinished()
    return self.finished
end

function ClassTimer:hasSlowmotiom()
    return self.slowmotion
end

function ClassTimer:getTimePassed()
    return self.TIME_LIMIT - self.time
end

function ClassTimer:getTimeLeft()
    return self.time
end

function ClassTimer:isActive()
    return self.active
end

function ClassTimer:update(dt)
    --[[
    Counts down the timer
    Parameters:
        dt - the time since the last frame
    Returns:
        <boolean> - whether the countdown has finished
    ]]--
    if not self.active or self.finished then
        return
    end

    --Timer is affect by slowmotion, so apply it
    if not self.realtime then
        dt = dt * bullettime.get()
    end

    self.time = self.time - dt

    if self.time <= 0 then
        self.finished = true
        self.active = false

        --Execute the function, if there is one
        if self.callback then
            self.callback()
        end
    end
end

function ClassTimer:reset()
    --[[
    Resets the instance back to its initial state
    ]]--
    self.time = self.TIME_LIMIT
    self.finished = false
    self.active = false
end

----------------------------------------------------------------------------------------------------
--Constructor
----------------------------------------------------------------------------------------------------

function ClassTimer:new(timeLimit, realtime, callback)
    --[[
    Attributes:
        timeLimit - the countdown to begin at
        realtime  - whether the timer runs at realtime or affected by slowmotion
        callback  - a function to execute when the timer is finished (can be nil to skip)
    ]]--
    errorHandler.assertNil(timeLimit)
    errorHandler.assertTrue(timeLimit > 0, "timeLimit must be > 0")

    local self = {}
    self.TIME_LIMIT = timeLimit
    self.time = self.TIME_LIMIT
    self.realtime = realtime
    self.callback = callback
    self.active = false
    self.finished = false

    setmetatable(self, {__index = ClassTimer})
    return self
end

----------------------------------------------------------------------------------------------------
--Interface
----------------------------------------------------------------------------------------------------

return ClassTimer