------------------------------------------------------------------------------------------------------------------
--Imports
----------------------------------------------------------------------------------------------------

local errorHandler = require("scenario/intersection/handlers/errorHandler")
local ClassPickup  = require("scenario/intersection/classes/pickups/classPickup")
local classHelper  = require("scenario/intersection/classHelper")
local soundHandler = require("scenario/intersection/handlers/soundHandler")

-------------------------------------------------------------------------------------------------------

local ClassCoin = classHelper.inheritsFrom(ClassPickup)

-------------------------------------------------------------------------------------------------------
--Static Attributes
----------------------------------------------------------------------------------------------------

ClassCoin.className = "ClassCoin"
ClassCoin.values = {
    BRONZE =  5000,
    SILVER = 10000,
    GOLD   = 25000
}

-------------------------------------------------------------------------------------------------------
--Static Methods
----------------------------------------------------------------------------------------------------

function ClassCoin:getTypes()
    --[[
    Returns:
        <table> - the list of coin types
    ]]--
    local keys = {}
    for k, _ in pairs(ClassCoin.values) do
        table.insert(keys, k:lower())
    end

    return keys
end

function ClassCoin:getValueOf(type)
    --[[
    Returns:
        <number> - the score for collecting this pickup
    ]]--
    errorHandler.assertNil(type)

    local value = ClassCoin.values[type:upper()]
    errorHandler.assertTrue(value ~= nil, type.." is not a known type")

    return value
end

----------------------------------------------------------------------------------------------------
--Overridden Methods
--See superclass for method descriptions
----------------------------------------------------------------------------------------------------

function ClassCoin:collect()
    ClassPickup.collect(self)
    soundHandler.play("collect_coin")
end

----------------------------------------------------------------------------------------------------
--Constructor
----------------------------------------------------------------------------------------------------

function ClassCoin:new(position, type)
    --[[
    See ClassPickup:new()
    ]]--
    errorHandler.assertValidElement(type, ClassCoin:getTypes(), type.." is an invalid type")

    local self = ClassPickup:new(position, type)
    self.value = ClassCoin:getValueOf(type)

    setmetatable(self, {__index = ClassCoin})
    return self
end

----------------------------------------------------------------------------------------------------
--Interface
----------------------------------------------------------------------------------------------------

return ClassCoin