----------------------------------------------------------------------------------------------------
--Imports
----------------------------------------------------------------------------------------------------

local errorHandler = require("scenario/intersection/handlers/errorHandler")
local ClassPickup  = require("scenario/intersection/classes/pickups/classPickup")
local classHelper  = require("scenario/intersection/classHelper")
local soundHandler = require("scenario/intersection/handlers/soundHandler")

------------------------------------------------------------------------------------------------------------------

local ClassMultiplier = classHelper.inheritsFrom(ClassPickup)

----------------------------------------------------------------------------------------------------
--Static Attributes
----------------------------------------------------------------------------------------------------

ClassMultiplier.className = "ClassMultiplier"
ClassMultiplier.values = {
    X2 = 2,
    X4 = 4,
    XH = 0.5
}

----------------------------------------------------------------------------------------------------
--Static Methods
----------------------------------------------------------------------------------------------------

function ClassMultiplier:getTypes()
    --[[
    Returns:
        <table> - the list of coin types
    ]]--
    local keys = {}
    for k, _ in pairs(ClassMultiplier.values) do
        table.insert(keys, k:lower())
    end

    return keys
end

function ClassMultiplier:getValueOf(type)
    --[[
    Returns:
        <number> - the score for collecting this pickup
    ]]--
    errorHandler.assertNil(type)

    local value = ClassMultiplier.values[type:upper()]
    errorHandler.assertTrue(value ~= nil, type.." is not a known type")

    return value
end

----------------------------------------------------------------------------------------------------
--Overridden Methods
--See superclass for method descriptions
----------------------------------------------------------------------------------------------------

function ClassMultiplier:collect()
    ClassPickup.collect(self)
    soundHandler.play("collect_multiplier_"..(self.type == "xh" and "negative" or "positive"))
end

----------------------------------------------------------------------------------------------------
--Constructor
----------------------------------------------------------------------------------------------------

function ClassMultiplier:new(position, type)
    --[[
    See ClassPickup:new()
    ]]--
    local self = ClassPickup:new(position, type)
    self.value = ClassMultiplier:getValueOf(type)

    setmetatable(self, {__index = ClassMultiplier})
    return self
end

------------------------------------------------------------------------------------------------------------------
--Interface
----------------------------------------------------------------------------------------------------

return ClassMultiplier