----------------------------------------------------------------------------------------------------
--Imports
----------------------------------------------------------------------------------------------------

local ClassAi      = require("scenario/intersection/classes/vehicles/classAi")
local classHelper  = require("scenario/intersection/classHelper")
local errorHandler = require("scenario/intersection/handlers/errorHandler")

----------------------------------------------------------------------------------------------------

local ClassTarget = classHelper.inheritsFrom(ClassAi)

-------------------------------------------------------------------------------------------------------
--Static Attributes
----------------------------------------------------------------------------------------------------

ClassTarget.className = "ClassTarget"
--Overridden values
ClassTarget.MAX_DAMAGE_SCORE = 15000
--[[Find the damage cut so that the same level of damage to an ai car and a target car results in
    their own max damage.
    Ie. maxScore = damage/cut where maxScore can equal different things but damage stays the same]]--
ClassTarget.DAMAGE_CUT = ClassTarget.MAX_DAMAGE_SCORE / (ClassAi.MAX_DAMAGE_SCORE / ClassAi.DAMAGE_CUT)

----------------------------------------------------------------------------------------------------
--Overridden Methods
--See superclass for method descriptions
----------------------------------------------------------------------------------------------------

function ClassTarget:update(dt)
    ClassAi.update(self, dt)

    --Place target sign above vehicle and rotate it
    local targetSignName = self.name.."_targetSign"
    TorqueScript.eval(targetSignName..'.position = "'..self.center:getX()..' '..self.center:getY()..' '..(self.center:getZ()+self.vehicleData.height)..'";')

    local anim = self.targetAnim
    anim.angle = anim.angle + anim.SPEED*dt
    TorqueScript.eval(targetSignName..'.rotation = "0 0 -1 '..anim.angle..'";')
end

function ClassTarget:setCrashed()
    ClassAi.setCrashed(self)

    guihooks.trigger("i_messages_enqueue", {
        text = "TARGET CAR TAKEDOWN",
        cssClass = "target",
        relPath = "target.png",
        duration = 5,
        colour = "Green"
    })

    self:changeSign("green")
end

function ClassTarget:reset()
    ClassAi.reset(self)

    --Reset sign
    self:changeSign("red")
end

----------------------------------------------------------------------------------------------------
--Instance Methods
----------------------------------------------------------------------------------------------------

function ClassTarget:changeSign(colour)
    --[[
    Changes the colour of the target sign
    Parameters:
        colour - the colour to change to (red or green)
    ]]--
    errorHandler.assertNil(colour)
    errorHandler.assertValidElement(colour, {"green", "red"}, colour.." is not a valid colour: can only be 'green' or 'red'")

    local targetSignName = self.name.."_targetSign"
    local signObj = scenetree.findObject(targetSignName)

    --Sign is already the correct colour so do nothing
    if signObj  then
        if signObj.shapeName:find(colour) then
            return
        end

        --Delete the existing object
        TorqueScript.eval(targetSignName..".delete();")
    end

    local modelName = colour == "red" and "target" or "target_crashed"

    local rot = self.obj:getRotation()
    --Recreate the new object
    TorqueScript.eval([[
        new TSStatic(]]..targetSignName..[[) {
            shapeName = "levels/smallgrid/art/intersection/vehicles/]]..modelName..[[.dae";
            dynamic = "1";
            scale = "0.25 0.25 0.25";
            position = "0 0 0";
            rotation = "]]..rot.x..[[ ]]..rot.y..[[ ]]..rot.z..[[ ]]..rot.w..[[";
            instanceColor = "0 0 0 0";
        };
    ]])
end

----------------------------------------------------------------------------------------------------
--Constructor
----------------------------------------------------------------------------------------------------

function ClassTarget:new(id, name, position, delay)
    --[[
    See ClassAi:new()
    ]]--
    local self = ClassAi:new(id, name, position, delay)

    self.targetAnim = {
        SPEED = 90,
        angle = 0
    }

    --Show lights
    self.obj:queueLuaCommand("electrics.set_lightbar_signal(1)");

    setmetatable(self, {__index = ClassTarget})

    self:changeSign("red")

    return self
end

----------------------------------------------------------------------------------------------------
--Interface
----------------------------------------------------------------------------------------------------

return ClassTarget
