angular.module('beamng.apps')
  .directive('intersectionCentralMessages', [function () {
    return {
      replace: true,
      template:
        '<div intersection-end-screen-root style="width:100%; height:100%; background-color:rgba(0, 0, 0, 0);">' +
        '    <link type="text/css" rel="stylesheet" href="modules/intersection/apps/style.css">' +
        '    <link type="text/css" rel="stylesheet" href="modules/intersection/apps/CentralMessages/app.css">' +
        '    <img class="textImage" id="image" alt="" src="modules/intersection/images/null.png" width="200px" height="200px">' +
        '    <div id="msgContainer" class="message"></div>' +
        '</div>',
      link: function (scope) {
        'use strict';

        scope.msgs = [];

        function resetValues() {
          scope.$evalAsync(function () {
            scope.msgs = [];
          });
        }

        /**
         * Handles text fading in and out by fading each character by a delay
         * @param {string}  text       - the text to set
         * @param {float}   timeLimit  - the total time the text takes to finish its animation
         * @param {boolean} isFadingIn - whether the fade is in or out
         * @param {string}  cssClass   - the class to apply to each character
         */
        function animateText(text, timeLimit, isFadingIn, cssClass) {
          let element = document.getElementById("msgContainer");
          let TIME_FACTOR = timeLimit/text.length;

          for (let i = 0; i < text.length; i++) {
            let time = i * TIME_FACTOR;
            let char = text.charAt(i);
            let id = "char_" + i;
            let span;

            if (isFadingIn) {
              if (document.getElementById(id))
                // Recreate it for centering
                document.getElementById(id).remove();

              // Create a new element for each char
              span = document.createElement("span");
              element.appendChild(span);

              span.className = cssClass;
              span.id = id;
              span.innerHTML = char === " " ? "&nbsp" : char;
              span.style.padding = "0px";
              span.style.fontSize = "60px";
              span.style.display = "inline-block";
            } else {
              // Simply get the element
              span = document.getElementById(id);
            }

            // Indicative of the scenario being reset
            if (!span)
              break;

            span.style.animation = "fade" + (isFadingIn ? "In" : "Out") + " " + time + "s forwards, fire 6s infinite linear";
          }
        }

        /**
         * Starts the next message, if there is one
         */
        function startNext() {
          if (scope.msgs.length > 0) {
            let message = scope.msgs[0];

            // Start the showing proceudres
            setText(message.text, message.durations, message.cssClass);
            setImage(message.relPath);
            animateImage(true, 1, 0, message.colour);
          }
        }

        /**
         * Sets the text using the default effects
         * @param {string} text       - the text to set
         * @param {Object} durations  - a set of durations for the fade in, middle (impasse) and fade out
         * @param {string} [cssClass] - the css class to use for this message (any animation(s) will be overridden)
         */
        function setText(text, durations, cssClass) {
          // Animate fade in
          animateText(text, durations.in, true, cssClass);

          setTimeout(function() {
            // Animate fade out
            animateText(text, durations.out, false, cssClass);
            // Animate out the image, even if one was not set
            animateImage(false, durations.out, 0);

            // Once finished, queue up the next message
            setTimeout(function () {
              scope.$apply(function () {
                scope.msgs.shift();
                startNext();
              });
            }, durations.out)
          }, (durations.impasse + durations.in)*1000);
        }

        /**
         * Fades the image in or out
         * @param {boolean} isFadingIn - whether the fade is in or out
         * @param {float}   duration   - the duration of the animation
         * @param {float}   delay      - the delay before the animation starts
         * @param {string}  colour     - the colour of the glow animation
         */
        function animateImage(isFadingIn, duration, delay, colour) {
          let img = document.getElementById("image");

          let fade = "fade" + (isFadingIn ? "In" : "Out") + "Partial " + duration + "s " + delay + "s forwards";
          let glow = "glow" + colour + " 1s infinite forwards";
          img.style.animation = fade + ", " + glow;
        }

        /**
         * Sets the image to go with the text
         * @param {string} relPath - the relative path (from ../../images) for the image to go with the text
         */
        function setImage(relPath) {
          let img = document.getElementById("image");
          img.src = "modules/intersection/images/" + relPath;
          img.style.opacity = "0";
        }

        /**
         * Queues a new message, comprising of text and an image
         * @param data - should contain:
                    text     - the text to show
                    cssClass - the css formatting class to apply to the text
                    relPath  - the relative path to the image to show as well
                    duration - the time the animation is dormant
                    colour   - the colour of the text glow
         */
        scope.$on('i_messages_enqueue', function (event, data) {
          let message = {
            text: data.text,
            cssClass: data.cssClass,
            relPath: data.relPath,
            // Animation will always take 1 second to fade each way
            durations: {
              "in": 1,
              "impasse": data.duration,
              "out": 1
            },
            colour: data.colour
          };

          // Queue it up
          scope.msgs.push(message);

          // If only this message exists, show the message right away
          if (scope.msgs.length === 1)
            startNext();
        });

        scope.$on('ScenarioResetTimer', resetValues);
      }
    };
  }]);
