scope = {};
scope.reset = false;
scope.speed = 1;

/**
 * Fades the timer
 * @param {boolean} out - whether to fade out or in the timer
 */
function fadeTimer(out) {
    let timer = document.getElementById("timer");

    if (out) {
        timer.style.animation = "fadeWait 2s linear forwards";
    } else {
        timer.style.animation = "none";
        timer.style.opacity = "1";
    }
}

/**
 * Sets the inner text on the center circle
 * @param {string} text - the text to set
 */
function setCenterText(text) {
    document.getElementById("timer").innerHTML = text;
}

/**
 * Sets the colour of the timer
 * @param {string} colour - the colour
 */
function setTimerColour(colour) {
    let timer = document.getElementById("timer");
    timer.style.backgroundColor = colour;
    timer.style.backgroundImage = ""
        + "linear-gradient(-90deg, transparent 50%, " + colour + " 50%), "
        + "linear-gradient(90deg, white 50%, transparent 50%)";

    // Create the shadow
    let shadow = colour + " 0 0 50px,".repeat(5);
    shadow = shadow.substring(0, shadow.length - 1);

    timer.style.textShadow = shadow;
    timer.style.webkitTextStroke = "thick " + colour;
}

/**
 * Plays the timer for a second
 * @param {string} baseColour - the colour to use as the timer bar
 * @param {string} text       - the text to display on completion of the timer
 */
function playTimer(baseColour, text) {
    return new Promise(function(resolve, reject) {
        let element = document.getElementById("timer");
        // setTimerColour(baseColour);

        let angle = 90;
        let colour = "white";

        // Timer plays at 60fps
        let timer = setInterval(function () {
            if (scope.reset) {
                return;
            }

            // Takes 1 second at 60fps to cycle
            angle += 360 / 60 * scope.speed;

            // Reached the bottom
            if (angle >= 270) {
                // Now at the top, so end
                if (colour === baseColour) {
                    clearInterval(timer);
                    setCenterText(text);

                    let sound;
                    if (text === "GO!")
                        sound = "countdown_go";
                    else if (text === "END")
                        sound = "countdown_end";
                    else
                        sound = "countdown_ready";

                    playSound(sound);

                    resolve();
                } else {
                    // At the bottom so flip colouring and restart
                    angle = angle - 180;
                    colour = baseColour;
                }
            }

            // element.style.backgroundImage = ""
            //     + "linear-gradient(" + angle + "deg, transparent 50%, " + colour + " 50%), "
            //     + "linear-gradient(90deg, white 50%, transparent 50%)";
        }, 1000 / 60);
    });
}

/**
 * Returns the rgb colour of the x% colour between red and green
 * @param {float} ratio - how far along the colour to work out is
 * @return {string} the rgba colour value
 */
function colourGradientRedGreen(ratio) {
    let r = ratio <= 0.5 ? 255 : 255 * 2 * (1 - ratio);
    let g = ratio >= 0.5 ? 255 : 255 * ratio * 2;

    return "rgb(" + r + ", " + g + ", 0)";
}

/**
 * Returns the rgb colour of the x% colour between red and green
 * @param {float} ratio - how far along the colour to work out is
 * @return {string} the rgba colour value
 */
function colourGradientYellowRed(ratio) {
    let g = 255 * (1 - ratio);
    return "rgb(255, " + g + ", 0)";
}

/**
 * Plays a sound triggered by an event
 * @param {string} event - the event name
 */
function playSound(event) {
    // bngApi.engineLua("require('scenario/intersection/handlers/soundHandler').play('" + event + "')");
}

/**
 * repeats the timer n times via recursion
 * @param {int}      totalRepeats - the number of times to repeat the timer
 * @param {int}      repeats      - the number of times the timer has already played
 * @param {string[]} msgs         - the messages to display at each timer
 * @param {function} colourFunc   - the function for determining colour based on timer value
 * @Return {Promise} the promise object
 */
function repeatTimer(totalRepeats, repeats, msgs, colourFunc) {
    let completionRate = repeats / totalRepeats;

    let colour = colourFunc(completionRate);

    if (repeats === totalRepeats) {
        setTimerColour(colour);
        return new Promise((resolve, _) => {
            if (scope.reset) {
                reject();
                return;
            }

            fadeTimer(true);
            resolve();
        }).catch(function() {
            // The last catch signals that the reset propagated successfully
            console.log("Reset successful");
        });
    }

    // The next iteration
    let iteration = new Promise(function (resolve, reject) {
        if (scope.reset) {
            reject();
            return;
        }

        playTimer(colour, msgs[0])
        .then(function() {
            console.log(true);
            resolve();
        });
    }).catch(endChain);

    // Add the next next iteration to this one
    return iteration.then(function () {
        if (scope.reset) {
            reject();
            return;
        }

        // Delete the message just shown
        msgs.shift();
        repeatTimer(totalRepeats, repeats + 1, msgs, colourFunc)
            .catch(endChain);
    }).catch(endChain);
}

/**
 * repeats the timer n times
 * uses a helper function to simplify the interface to this function
 * @param {int}      totalRepeats - the number of times to repeat the timer
 * @param {string[]} msgs         - the messages to display at each timer
 * @param {function} colourFunc   - the function for determining colour based on timer value
 */
function playerNTimers(totalRepeats, msgs, colourFunc) {
    playSound("countdown_ready");

    repeatTimer(totalRepeats, 0, msgs, colourFunc)
        .catch(endChain);
}

/**
 * The end chain rejection promise
 * @returns {Promise} the promise rejection
 */
function endChain() {
    return Promise.reject("Reset");
}

scope.speed = 3;
playerNTimers(3, ["A", "B", "C"], function(_) {
    return "rgb(255,255,255)";
});