angular.module('beamng.apps')
  .directive('intersectionEndScreen', ['bngApi', 'gamepadNav', function (bngApi) {
    return {
      templateUrl: 'modules/intersection/apps/EndScreen/templateBMNG.html',
      replace: true,
      link: function (scope) {
        'use strict';

        scope.reset = false;

        // Length of the medal bar element
        let MEDAL_UI_LENGTH = 820;
        // The width of the image / 2
        let MEDAL_OFFSET = 53 / 2;

        let POINTS_FACTOR = 0;

        let ROW_COLOURS = {
          "success": "#00FF00",
          "failure": "red",
          "perfect": "gold",
          "pickup" : "orange"
        };

        let TOTAL_ROW_ID = 0;

        /**
         *Creates a row at the end of the data table
         *@param {int} id - the id of the newly created row
         */
        function createRowAtEnd(id) {
          let newRow = document.getElementById("pointsToAdd").insertRow(-1);
          newRow.id = "row_" + id;

          for (let i = 0; i < 2; i++) {
            let cell = newRow.insertCell(0);
            cell.innerHTML = "&nbsp";
          }
        }

        /**
         * Updates the total row
         * @param {int} value - the value to set
         * @param {String} changeReason - the reason the update is happening
         */
        function updateTotalRow(value, changeReason) {
          let animation = "";
          if (changeReason === "negative") {
            animation = "glowRed 0.5s";
          } else if (changeReason === "positive") {
            animation = "glowGreen 0.5s";
          } else if (changeReason === "neutral") {
            animation = "glowBlue 0.5s";
          } else if (changeReason === "pickup") {
            animation = "addMultiplier 1s";
          } else {
            console.log("Total updated with unknown reasoning [" + changeReason + "]");
          }

          let row = document.getElementById("row_" + TOTAL_ROW_ID);

          if (!row)
            return;

          let cells = row.cells;
          let left = cells[0];
          let right = cells[1];

          // Add bold tags + animation
          left.innerHTML = "<span style='animation:" + animation + "; font-size:32px; color:white;'><b>Total</b></span>";
          right.innerHTML = "<span style='animation:" + animation + "; font-size:32px; color:white;'><b>" + value + "</b></span>";
        }

        /**
         * Updates the a row's cells in the results table
         * @param id     - the id of the newly created row
         * @param reason - the reason for the points awarded (left cell)
         * @param value  - the amount of points awarded (right cell)
         * @param [max]  - the maximum value that can be awarded
         */
        function updateRow(id, reason, value, max) {
          let row = document.getElementById("row_" + id);
          if (!row) {
            return;
          }

          let valueType = "success";
          let cls = "";

          if (max && max === value) {
            valueType = "perfect";
            cls = "perfect";
            playSound("points_perfect")
          } else if (value < 0) {
            valueType = "failure";
          } else if (reason === "pickup") {
            valueType = "pickup"
          }

          let cells = row.cells;
          let colour = ROW_COLOURS[valueType];

          cells[0].innerHTML = "<span style='color:" + colour + ";'><b>" + reason + "</span>";
          cells[1].innerHTML = "<span class='" + cls + "' style='color:" + colour + ";'><b>" + value + "</span>";
        }

        /**
         * Adds the pickups to the table and applies them to the score
         * @param purpose      - the pickup purpose (ie. Coins, Multiplier, etc.)
         * @param pickups      - an array of pickups gotten
         * @param currentTotal - where the total left off from accumulating
         * @param rowId        - the id of the pickup row in the table
         * @param changeFunc   - the function to change the total with
         * @return {Promise} - the promise object
         */
        function addPickups(purpose, pickups, currentTotal, rowId, changeFunc) {
          return new Promise(function (resolve, reject) {
            if (!pickups) {
              resolve(currentTotal);
              return;
            }

            let row = document.getElementById("row_" + rowId);
            if (!row) {
              reject(Error("The rowId=" + rowId + " does not correspond with any row"));
              return;
            }

            let cells = row.cells;

            cells[0].innerHTML = "<span style='color:orange'><b>" + purpose + "</b></span>";
            cells[1].innerHTML = "";

            // Let the row be added but do not add in anything for this row
            if (pickups.isEmpty()) {
              resolve(currentTotal);
              return;
            }

            let index = 0;
            let timer = setInterval(function () {
              if (scope.reset) {
                reject("Reset");
                return;
              }

              let current = pickups[index];
              let name = current.type;
              let value = current.value;
              cells[1].innerHTML += "<img src='modules/intersection/images/pickups/" + name + ".png' style='height:28px; width:28px;' alt=''>";

              playSound("collect_coin");

              currentTotal = changeFunc(currentTotal, value);
              // Total row is 2 rows down from the multiplier row
              updateTotalRow(currentTotal, "pickup");

              index++;

              if (index >= pickups.length) {
                clearInterval(timer);
                resolve(currentTotal);
              }
            }, 500);
          });
        }

        /**
         * Fills the points table with those awarded
         * @param {Object} data - the array of (point, reason) pairs to add to the table
         * @return {Promise} - the promise object
         */
        function fillPoints(data) {
          return new Promise(function (resolve, reject) {
            if (scope.reset) {
              reject("Reset");
              return;
            }

            let multipliers = null;
            let multiRow = null;
            let coins = null;
            let coinRow = null;

            if (true) {
              // Create rows beforehand
              for (let i = 0; i < data.length; i++) {
                createRowAtEnd(i);
                if (i === data.length - 1) {
                  // Create blank row between breakdown and total
                  createRowAtEnd(i + 1);
                }
              }

              let index = 0;
              let total = 0;

              let timer = setInterval(function () {
                let pts = data[index];
                let reason = pts.reason;
                let amount = pts.amount;
                let max = pts.max;

                if (reason === "Multipliers") {
                  multipliers = amount;
                  multiRow = index;
                } else if (reason === "Coins") {
                  coins = amount;
                  coinRow = index;
                } else {
                  playSound("points");

                  updateRow(index, reason, amount, max);

                  let reasoning = "";

                  // Determine reason for total update changes
                  if (amount > 0) {
                    reasoning = "positive";
                  } else if (amount < 0) {
                    reasoning = "negative";
                  } else if (amount === 0) {
                    reasoning = "neutral";
                  }

                  total += amount;

                  updateTotalRow(total, reasoning);
                }

                index++;

                // Ignore last element as it will be the total row, which is not relevant
                if (index === data.length - 1) {
                  clearInterval(timer);

                  //  Make sure that total does not finish below 0
                  if (total < 0) {
                    total = 0;
                    updateTotalRow(0, "pickup");
                  }

                  resolve([multipliers, multiRow, coins, coinRow, total]);
                }
              }, 600);
            }
//          else {
//            // Create a row to explain why user failed
//            createRowAtEnd(0);
//            updateRow(0, "Failure", reason);
//            wait();
//          }
          });
        }

        /**
         * Shows the target vehicle and whether it was achieved or not, adjusting score as necessary
         */
        function showTarget(targetData, total) {
          return new Promise(function(resolve, reject) {
            // Nothing to do so resolve early
            if (!targetData || !targetData.gotten) {
              resolve(total);
              return;
            }

            if (scope.reset) {
              reject("Reset");
              return;
            }


            let target = document.getElementById("target");
            target.style.animation = "targetCrashed 1s, pulseX4 2s infinite forwards";

            total += targetData.value;
            updateTotalRow(total, "pickup");

            playSound("target_crashed");

            resolve(total);
          });
        }

        /**
         * Adds in the multipliers
         * @param [multipliers] - the multiplier data
         * @param total         - the current total to work with
         * @return {Promise} - the promise object, with the new total
         */
        function addMultiplier(multipliers, total) {
          return new Promise(function (resolve, reject) {
            // Nothing to do so resolve early
            if (!multipliers) {
              resolve(total);
              return;
            }

            let multiImg = document.getElementById("multiplier");
            let multiArrow = document.getElementById("multiArrow");
            let index = 0;

            let timer = setInterval(function () {
              if (scope.reset) {
                reject("Reset");
                return;
              }

              let multi = multipliers[index];
              let type = multi.type;

              let downgrade = type === "xh";
              let soundEvent = "collect_multiplier_" + (downgrade ? "negative" : "positive");

              if (multiImg.src.match("null")) {
                // Show initial image animation
                multiImg.style.animation = "multiplierRotate 1.25s forwards";
              } else {
                // Show upgrade/downgrade arrow
                let change = downgrade ? "downgrade" : "upgrade";
                multiArrow.src = "modules/intersection/images/" + change + ".png";
                multiArrow.style.animation = change + " 1.25s forwards";

                // Animates the change
                multiImg.style.animation = "multiplier" + change[0].toUpperCase() + change.slice(1) + " 1.25s forwards";
              }

              playSound(soundEvent);

              // Set source file
              multiImg.src = "modules/intersection/images/pickups/" + type + ".png";

              index++;

              if (index >= multipliers.length) {
                multiImg.style.animation += " , pulse" + type.toUpperCase() + " 2s infinite forwards";

                // Chosen multiple will always be the last
                let multiplier = multi.value;
                // Display red as loss of score for half-multiplier
                let change = multiplier === 0.5 ? "negative" : "pickup";

                // Handle non-integer multipliers showing fractional scores
                total = Math.floor(total * multiplier);

                updateTotalRow(Math.floor(total), change);

                clearInterval(timer);
                resolve(total);
              }
            }, 1500);
          });
        }

        /**
         * Shows the personal best banner
         */
        function showPB() {
          document.getElementById("pbBox").style.display = "block";
        }

        /**
         * Shows the bottom buttons
         */
        function showButtons() {
          document.getElementById("buttons").style.display = "block";
          document.getElementById("retry").addEventListener("click", restartScenario, false);

          // let filePath = "levels/smallgridspace/scenarios/cargames/roam/cargameshub.json";
          // document.getElementById("hubWorld").addEventListener("click", function () {
          //   loadScenario(filePath);
          // });

//          if (data.scenarioData.nextScenario) {
//            document.getElementById("nextScenario").addEventListener("click", function() {loadScenario(data.scenarioData.nextScenario);})
//          } else {
//           document.getElementById("nextScenario").style.display = "none";
//          }
        }

        /**
         * Shows the medal achieved
         * @param {String}  medalAchieved - the medal achieved
         * @param {boolean} betterMedal   - whether this is a better medal than previous
         * @param {boolean} betterScore   - whether this is a better score than previous
         */
        function showMedal(medalAchieved, betterMedal, betterScore) {
          if (medalAchieved === "") {
            medalAchieved = "none";
          }

          let medalElement = document.getElementById("medal");
          medalElement.style.display = "block";
          medalElement.src = "modules/intersection/images/medals/" + medalAchieved + ".png";
          medalElement.style.animation = "glow 2s infinite forwards, fadeInScale 1.5s";

          if (medalAchieved === "dev") {
            // Add in medal effects
            document.getElementById("medalWrapper").style.animation = "pulseX4 2s infinite forwards";
            // Add in counter effects
            document.getElementById("playerScore").style.backgroundColor = "rgb(0,255,0)";
            document.getElementById("scoreCounter").style.animation = "pulseX4 2s infinite forwards";

            document.getElementById("devBox").style.display = "block";
          }

          if (betterMedal) {
            medalElement.addEventListener("animationend", function () {
              let element = document.getElementById("effect");
              element.style.display = "block";
              element.style.animation = "shimmerBackground 2s linear infinite";

              // let colour = medalToColour(medalAchieved);
              // effect.style.backgroundImage = "linear-gradient(to right, " + colour + "00 0%," + colour + "00 50%," + colour + "21 77%," + colour + "F3 92%," + colour + "00 100%)";
            });
          }

          if (betterScore)
            showPB();

          showButtons();
        }

        /**
         * Shows the player's score compared to the medals'
         * @param {int} total - the total amount of points accumulated
         * @param {Object} medalInfo - the medal data, with key pairs (score,gotten)
         * @return {Promise} - the promise object
         */
        function showRanking(total, medalInfo) {
          return new Promise(function (resolve, reject) {
            // Ignore since nothing will be shown
            if (total === 0) {
              resolve("", false);
              return;
            }

            //The total time it should take to animate the timer
            const REAL_TIME = 4;

            let incrementMultiplier = 0;
            let t = 0;
            let currentPos = 0;
            let pendEnd = false;

            let medalAchieved = "";
            let betterScore = false;

            // The position this score will go to
            let totalPos = POINTS_FACTOR * total;

            playSound("points_final");

            if (true) {
              let timer = setInterval(function () {
                if (scope.reset) {
                  reject("Reset");
                  return;
                }

                //Since this function updates every 10ms
                t += 1 / 100;

                //f(t) = (t/K - 1)³ + 1
                incrementMultiplier = Math.pow((t / REAL_TIME - 1), 3) + 1;

                //Account for x³ approaching 1 unlikely to ever be 1 exactly
                if (Math.abs(incrementMultiplier - 1) <= Math.pow(10, -6)) {
                  incrementMultiplier = 1;
                  pendEnd = true;
                }

                currentPos = Math.min(totalPos, MEDAL_UI_LENGTH) * incrementMultiplier;

                for (let [key, data] of Object.entries(medalInfo)) {
                  // dev medal does not have this attribute
                  if (data.pos) {
                    if (currentPos >= data.pos && key !== "best") {
                      document.getElementById(key + "_line").style.border = "2px solid black";
                      // Change bar colour
                      document.getElementById("playerScore").style.backgroundColor = medalToColour(key);

                      if (!data.gotten) {
                        // Animate collection
                        document.getElementById("medal_" + key).style.animation = "fadeOut 2s forwards";
                      }

                      medalAchieved = key;

                      // Remove the key as this medal has been achieved
                      delete medalInfo[key];
                      // Better than best means greater than, unlike medals which is inclusive of their target
                    } else if (currentPos > data.pos) {
                      betterScore = true;
                      delete medalInfo[key];
                    }
                  }
                }

                document.getElementById("playerScore").style.width = currentPos + "px";

                if (pendEnd) {
                  clearInterval(timer);

                  // Check after as does not appear on the counter UI
                  if (total >= medalInfo.dev.score) {
                    medalAchieved = "dev";
                  }

                  playSound(medalAchieved === "" ? "medal_none" : "medal_" + medalAchieved);

                  resolve(medalAchieved, betterScore);
                }
              }, 10);
            } else {
              // noinspection UnreachableCodeJS
              resolve(medalAchieved, betterScore);
            }
          });
        }

        /**
         * Toggles the UI from view
         */
        function toggleUI() {
          let info = document.getElementById("info");
          let button = document.getElementById("toggle");

          if (info.style.visibility === "visible") {
            info.style.visibility = "hidden";

            button.innerHTML = "&nbsp;show&nbsp;";
            button.className = "show";
          } else {
            info.style.visibility = "visible";

            button.innerHTML = "&nbsp;hide&nbsp;";
            button.className = "hide";
          }
        }

        /**
         * Returns the specific colour of a medal
         * @param {String} medal - the name of the medal
         * @return {String} - the medal colour string
         */
        function medalToColour(medal) {
          medal = medal.toLowerCase();

          if (medal === "bronze") {
            return "#CD7F32";
          } else if (medal === "silver") {
            return "#C0C0C0";
          } else if (medal === "gold") {
            return "#FFD700";
          } else if (medal === "chaos") {
            return "#B9F2FF";
          } else if (medal === "best") {
            return "#00ff00";
          } else {
            return "#FF0000";
          }
        }

        /**
         * Restarts the current scenario
         */
        function restartScenario() {
          bngApi.engineLua("scenario_scenarios.restartScenario()");
        }

        /**
         * Plays a sound triggered by an event
         * @param event - the event name
         */
        function playSound(event) {
          bngApi.engineLua("require('scenario/intersection/handlers/soundHandler').play('" + event + "')");
        }

        /**
         * Loads a given scenario
         */
        function loadScenario(filePath) {
          filePath = bngApi.serializeToLua(filePath);
          let cmd = "scenario_scenariosLoader.start(scenario_scenariosLoader.loadScenario(" + filePath + ", nil, " + filePath + "))";
          bngApi.engineLua(cmd);
        }

        /**
         * Places the medals on the UI
         * @param {Object} medalScores - key pairs of (score,gotten)
         * @return {Object} - the medalScores array with a new attribute for each medal: pos
         */
        function setupRanking(medalScores) {
          for (let [key, data] of Object.entries(medalScores)) {
            // Dev is a secret medal so don't show
            if (key === "dev")
              continue;

            // Center the medal
            let pos = POINTS_FACTOR * data.score - MEDAL_OFFSET;

            let imgHTML = "";
            if (!data.gotten && key !== "best") {
              imgHTML = "<div id='" + key + "_wrapper'>"
                + "<img id='medal_" + key + "' src='modules/intersection/images/medals/" + key + ".png' style='width:" + MEDAL_OFFSET * 2 + "px; height:50px; opacity:1; z-index:1; position:absolute; top:5px; left:" + pos + "px;' alt=''>"
                + "<span id='" + key + "_line' style='border:2px solid " + medalToColour(key) + "; height:80px; position:absolute; top:-10px; left:" + (pos + MEDAL_OFFSET - 2) + "px;'></span></img></div>";
            } else {
              imgHTML = "<div id='" + key + "_wrapper'><span id='" + key + "_line' style='border:2px solid " + medalToColour(key) + "; height:80px; position:absolute; top:-10px; left:" + (pos + MEDAL_OFFSET - 2) + "px;'></span></div>";
            }

            document.getElementById("scoreCounter").innerHTML += imgHTML;

            // Change to array and add pos to it
            medalScores[key].pos = pos + MEDAL_OFFSET;
          }

          return medalScores;
        }

        /**
         * Sets up the end screen by animating the opening and adjusting colour schemes and text
         * @param {String} scenarioTitle - the name of the scenario
         * @param {Object} targetData    - the target vehicle data
         * @returns {Promise} - the promise object
         */
        function setupEndScreen(scenarioTitle, targetData) {
          let title = document.getElementById("scenarioTitle");
          title.innerHTML = "<b>" + scenarioTitle.toUpperCase() + "</b>";

//          if (!success) {
//            //Remove outline as this will be empty
//            document.getElementById("scoreCounter").style.border = "";
//          }

          //Animate finish div
          let finish = document.getElementById("finish");
          finish.style.display = "block";
          finish.style.animation = "fire 8s linear infinite, slideDownOut 3s linear forwards";

          //Animate results div after the finish div
          let info = document.getElementById("info");
          info.style.display = "block";
          info.style.animation = "slideDownPause 1.75s linear 2.5s forwards";

          //Animate the button to do with the results div
          let button = document.getElementById("toggle");
          button.style.display = "block";
          button.style.animation = "slideDownPauseButton 1.75s linear 2.5s forwards";

          // Setup target image if necessary
          if (targetData) {
            document.getElementById("target").src = "modules/intersection/images/target.png";
          }

          return new Promise(function (resolve, reject) {
            if (scope.reset) {
              reject("Reset");
              return;
            }

            setTimeout(function () {
              document.getElementById("toggle").addEventListener("click", toggleUI, false);
              resolve();
            }, 4250);
          });
        }

        /**
         * The end chain rejection promise
         * @returns {Promise} - the promise rejection
         */
        function end() {
          return Promise.reject("Reset");
        }

        /**
         * Displays the end screen
         * @param {Object} data - the data
         *                 {boolean} data.success   - whether the scenario was successful
         *                 {String}  data.reason    - the reason the scenario failed
         *                 {Object}  data.medalData - information on the medals and previous best
         *                 {String}  data.name      - the name of the scenario
         *                 {Object}  data.results   - the points breakdown
         */
        scope.$on("i_endScreen", function (event, data) {
          scope.reset = false;

          let success = data.success;
          // let reason = data.reason;

          let scenarioTitle = data.name;

          let medalData = data.medalData;

          POINTS_FACTOR = success
            // The ratio of score to pixels - *1.05 for gap past diamond/best
            ? MEDAL_UI_LENGTH / (Math.max(medalData.chaos.score, medalData.best.score) * 1.05)
            : 0;

          let finish = document.getElementById("finish");

          if (!success) {
            //        finish.style.webkitTextStrokeColor = "lightred";
            //        finish.style.color = "red";
            //        finish.innerHTML = "FAIL";
            //        document.getElementById("medal").style.display = "none";
            //
            //        start();
          } else {
            finish.innerHTML = "FINISHED";

            let results = data.results;
            let multipliers = data.multipliers;
            let target = data.target;

            TOTAL_ROW_ID = results.length;

            // Last index will be total
            let total = results[results.length - 1].amount;
            let betterMedal = medalData.best.newMedal;

            let medalInfo = setupRanking(medalData);

            /**
             * Waits for a given amount of time
             * @param {int} n - the time (in ms) to wait for
             * @returns {Promise} - the promise object
             */
            let wait = function(n) {
              return new Promise(function (resolve, reject) {
                if (scope.reset) {
                  reject("Reset");
                  return;
                }

                setTimeout(() => resolve(), n);
              })
            };

            let multiRowId = null;
            let multiRow = null;
            let coinRowId = null;
            let coinsRow = null;
            let currentTotal = null;
            // Execute the chain of functions

            // 1 - setup endscreen
            setupEndScreen(scenarioTitle, target)
              .catch(end)
              .then(function () {
                // 2 - fill in points allocation
                return fillPoints(results);
              })
              .catch(end)
              .then(function (result) {
                // 2.1 - wait
                // Store the results from fillPoints()
                multiRow = result[0];
                multiRowId = result[1];
                coinsRow = result[2];
                coinRowId = result[3];
                currentTotal = result[4];
                return wait(250);
              })
              .catch(end)
              .then(function () {
                // 3 - add in the multiplier
                return addMultiplier(multipliers, currentTotal);
              })
              .catch(end)
              .then(function (newTotal) {
                // 3.1 - wait
                currentTotal = newTotal;
                return wait(1000);
              })
              .catch(end)
              .then(function () {
                // 4 - add in the pickups
                return addPickups("Coins", coinsRow, currentTotal, coinRowId, (t, a) => t + a);
              })
              .catch(end)
              .then(function (newTotal) {
                // 4.1 - wait
                currentTotal = newTotal;
                return wait(1000);
              })
              .catch(end)
              .then(function() {
                // 5 - show the target
                return showTarget(target, currentTotal);
              })
              .catch(end)
              .then(function (newTotal) {
                // 5.1 - wait
                currentTotal = newTotal;
                return wait(1000);
              })
              .catch(end)
              .then(function () {
                // 6 - show the points-medal bar
                return showRanking(total, medalInfo);
              })
              .catch(end)
              .then(function (medalAchieved, betterScore) {
                // 7 - show the medal achieved
                showMedal(medalAchieved, betterMedal, betterScore);
              })
              .catch(function() {console.log("Reset successful")});
          }
        });

        scope.$on("i_reset", function () {
          scope.reset = true;
        });
      }
    };
  }]);
